@@ -39,7 +39,7 @@ from .output_widget import OutputWidget
 from .util import ensure_async, run_hook, run_sync
 
 
-def timestamp(msg: t.Optional[Dict] = None) -> str:
+def timestamp(msg: t.Optional[t.Dict] = None) -> str:
     if msg and 'header' in msg:  # The test mocks don't provide a header, so tolerate that
         msg_header = msg['header']
         if 'date' in msg_header and isinstance(msg_header['date'], datetime.datetime):
@@ -106,6 +106,25 @@ class NotebookClient(LoggingConfigurable):
         ),
     ).tag(config=True)
 
+    error_on_timeout: t.Optional[t.Dict] = Dict(
+        default_value=None,
+        allow_none=True,
+        help=dedent(
+            """
+            If a cell execution was interrupted after a timeout, don't wait for
+            the execute_reply from the kernel (e.g. KeyboardInterrupt error).
+            Instead, return an execute_reply with the given error, which should
+            be of the following form::
+
+                {
+                    'ename': str,  # Exception name, as a string
+                    'evalue': str,  # Exception value, as a string
+                    'traceback': list(str),  # traceback frames, as strings
+                }
+            """
+        ),
+    ).tag(config=True)
+
     startup_timeout: int = Integer(
         60,
         help=dedent(
@@ -269,7 +288,9 @@ class NotebookClient(LoggingConfigurable):
         """,
     ).tag(config=True)
 
-    kernel_manager_class: KernelManager = Type(config=True, help='The kernel manager class to use.')
+    kernel_manager_class = Type(
+        config=True, klass=KernelManager, help='The kernel manager class to use.'
+    )
 
     on_notebook_start: t.Optional[t.Callable] = Callable(
         default_value=None,
@@ -340,6 +361,18 @@ class NotebookClient(LoggingConfigurable):
         ),
     ).tag(config=True)
 
+    on_cell_executed: t.Optional[t.Callable] = Callable(
+        default_value=None,
+        allow_none=True,
+        help=dedent(
+            """
+            A callable which executes just after a code cell is executed, whether
+            or not it results in an error.
+            Called with kwargs ``cell``, ``cell_index`` and ``execute_reply``.
+            """
+        ),
+    ).tag(config=True)
+
     on_cell_error: t.Optional[t.Callable] = Callable(
         default_value=None,
         allow_none=True,
@@ -347,7 +380,7 @@ class NotebookClient(LoggingConfigurable):
             """
             A callable which executes when a cell execution results in an error.
             This is executed even if errors are suppressed with ``cell_allows_errors``.
-            Called with kwargs ``cell` and ``cell_index``.
+            Called with kwargs ``cell`, ``cell_index`` and ``execute_reply``.
             """
         ),
     ).tag(config=True)
@@ -359,7 +392,7 @@ class NotebookClient(LoggingConfigurable):
 
         return AsyncKernelManager
 
-    _display_id_map: t.Dict[str, t.Dict] = Dict(
+    _display_id_map = Dict(
         help=dedent(
             """
               mapping of locations of outputs with a given display_id
@@ -392,7 +425,7 @@ class NotebookClient(LoggingConfigurable):
             """,
     ).tag(config=True)
 
-    resources: t.Dict = Dict(
+    resources = Dict(
         help=dedent(
             """
             Additional resources used in the conversion process. For example,
@@ -526,11 +559,16 @@ class NotebookClient(LoggingConfigurable):
             Kernel client as created by the kernel manager ``km``.
         """
         assert self.km is not None
-        self.kc = self.km.client()
-        await ensure_async(self.kc.start_channels())  # type:ignore[func-returns-value]
         try:
-            await ensure_async(self.kc.wait_for_ready(timeout=self.startup_timeout))
-        except RuntimeError:
+            self.kc = self.km.client()
+            await ensure_async(self.kc.start_channels())  # type:ignore[func-returns-value]
+            await ensure_async(self.kc.wait_for_ready(timeout=self.startup_timeout))  # type:ignore
+        except Exception as e:
+            self.log.error(
+                "Error occurred while starting new kernel client for kernel {}: {}".format(
+                    self.km.kernel_id, str(e)
+                )
+            )
             await self._async_cleanup_kernel()
             raise
         self.kc.allow_stdin = False
@@ -603,7 +641,10 @@ class NotebookClient(LoggingConfigurable):
 
         if not self.km.has_kernel:
             await self.async_start_new_kernel(**kwargs)
+
+        if self.kc is None:
             await self.async_start_new_kernel_client()
+
         try:
             yield
         except RuntimeError as e:
@@ -720,14 +761,20 @@ class NotebookClient(LoggingConfigurable):
         task_poll_kernel_alive: asyncio.Future,
     ) -> t.Dict:
 
+        msg: t.Dict
         assert self.kc is not None
         new_timeout: t.Optional[float] = None
         if timeout is not None:
             deadline = monotonic() + timeout
             new_timeout = float(timeout)
+        error_on_timeout_execute_reply = None
         while True:
             try:
-                msg: t.Dict = await ensure_async(self.kc.shell_channel.get_msg(timeout=new_timeout))
+                if error_on_timeout_execute_reply:
+                    msg = error_on_timeout_execute_reply
+                    msg['parent_header'] = {'msg_id': msg_id}
+                else:
+                    msg = await ensure_async(self.kc.shell_channel.get_msg(timeout=new_timeout))
                 if msg['parent_header'].get('msg_id') == msg_id:
                     if self.record_timing:
                         cell['metadata']['execution']['shell.execute_reply'] = timestamp(msg)
@@ -751,7 +798,7 @@ class NotebookClient(LoggingConfigurable):
                 assert timeout is not None
                 task_poll_kernel_alive.cancel()
                 await self._async_check_alive()
-                await self._async_handle_timeout(timeout, cell)
+                error_on_timeout_execute_reply = await self._async_handle_timeout(timeout, cell)
 
     async def _async_poll_output_msg(
         self, parent_msg_id: str, cell: NotebookNode, cell_index: int
@@ -790,21 +837,26 @@ class NotebookClient(LoggingConfigurable):
 
     async def _async_handle_timeout(
         self, timeout: int, cell: t.Optional[NotebookNode] = None
-    ) -> None:
+    ) -> t.Union[None, t.Dict]:
 
         self.log.error("Timeout waiting for execute reply (%is)." % timeout)
         if self.interrupt_on_timeout:
             self.log.error("Interrupting kernel")
             assert self.km is not None
             await ensure_async(self.km.interrupt_kernel())
+            if self.error_on_timeout:
+                execute_reply = {"content": {**self.error_on_timeout, "status": "error"}}
+                return execute_reply
+            return None
         else:
+            assert cell is not None
             raise CellTimeoutError.error_from_timeout_and_cell(
                 "Cell execution timed out", timeout, cell
             )
 
     async def _async_check_alive(self) -> None:
         assert self.kc is not None
-        if not await ensure_async(self.kc.is_alive()):
+        if not await ensure_async(self.kc.is_alive()):  # type:ignore
             self.log.error("Kernel died while waiting for execute reply.")
             raise DeadKernelError("Kernel died")
 
@@ -825,7 +877,7 @@ class NotebookClient(LoggingConfigurable):
                 await self._async_check_alive()
                 cummulative_time += self.shell_timeout_interval
                 if timeout and cummulative_time > timeout:
-                    await self._async_async_handle_timeout(timeout, cell)
+                    await self._async_handle_timeout(timeout, cell)
                     break
             else:
                 if msg['parent_header'].get('msg_id') == msg_id:
@@ -857,7 +909,9 @@ class NotebookClient(LoggingConfigurable):
             or exec_reply_content.get('ename') in self.allow_error_names
             or "raises-exception" in cell.metadata.get("tags", [])
         )
-        await run_hook(self.on_cell_error, cell=cell, cell_index=cell_index)
+        await run_hook(
+            self.on_cell_error, cell=cell, cell_index=cell_index, execute_reply=exec_reply
+        )
         if not cell_allows_errors:
             raise CellExecutionError.from_cell_and_msg(cell, exec_reply_content)
 
@@ -962,6 +1016,9 @@ class NotebookClient(LoggingConfigurable):
 
         if execution_count:
             cell['execution_count'] = execution_count
+        await run_hook(
+            self.on_cell_executed, cell=cell, cell_index=cell_index, execute_reply=exec_reply
+        )
         await self._check_raise_for_error(cell, cell_index, exec_reply)
         self.nb['cells'][cell_index] = cell
         return cell
@@ -970,7 +1027,7 @@ class NotebookClient(LoggingConfigurable):
 
     def process_message(
         self, msg: t.Dict, cell: NotebookNode, cell_index: int
-    ) -> t.Optional[t.List]:
+    ) -> t.Optional[NotebookNode]:
         """
         Processes a kernel message, updates cell state, and returns the
         resulting output object that was appended to cell.outputs.
@@ -988,7 +1045,7 @@ class NotebookClient(LoggingConfigurable):
 
         Returns
         -------
-        output : dict
+        output : NotebookNode
             The execution output payload (or None for no output).
 
         Raises
@@ -1037,6 +1094,7 @@ class NotebookClient(LoggingConfigurable):
     ) -> t.Optional[NotebookNode]:
 
         msg_type = msg['msg_type']
+        out = None
 
         parent_msg_id = msg['parent_header'].get('msg_id')
         if self.output_hook_stack[parent_msg_id]:
@@ -1067,7 +1125,7 @@ class NotebookClient(LoggingConfigurable):
 
         outs.append(out)
 
-        return out
+        return out  # type:ignore[no-any-return]
 
     def clear_output(self, outs: t.List, msg: t.Dict, cell_index: int) -> None:
 
